/*
  KPUInput - KeePass Plugin
  Copyright (C) 2020-2022 Dominik Reichl <dominik.reichl@t-online.de>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <cstdlib>
#include <cstddef>
#include <cstdint>
#include <cstdio>
#include <cstring>
#include <exception>
#include <string>
#include <unordered_map>
#include <unordered_set>

#include <fcntl.h>
#include <time.h>
#include <unistd.h>

#include <linux/input.h>
#include <linux/uinput.h>

#include <systemd/sd-bus.h>
#include <systemd/sd-daemon.h>

#include <xkbcommon/xkbcommon.h>

#ifndef DLL_PUBLIC
#if defined(__GNUC__)
#define DLL_PUBLIC __attribute__((visibility("default")))
#elif defined(_WIN32)
#define DLL_PUBLIC __declspec(dllexport)
#endif
#endif

#define DLL_PUBLIC_C extern "C" DLL_PUBLIC

#define KPUI_K_CODE        0x00FFFF
#define KPUI_K_MOD_SHIFT_L 0x010000
#define KPUI_K_MOD_SHIFT_R 0x020000
#define KPUI_K_MOD_CTRL_L  0x040000
#define KPUI_K_MOD_CTRL_R  0x080000
#define KPUI_K_MOD_ALT_L   0x100000
#define KPUI_K_MOD_ALT_R   0x200000

#define KPUI_S_DOWN 1
#define KPUI_S_UP   2

using namespace std;

// Not thread-safe; sd_bus is associated with the calling thread
class CSdBus
{
public:
	CSdBus() : m_pBus(nullptr)
	{
		if(sd_booted() <= 0) return;

		sd_bus_default_system(&m_pBus);
	}

	virtual ~CSdBus()
	{
		if(m_pBus == nullptr) return;

		sd_bus_flush_close_unref(m_pBus);
		m_pBus = nullptr;
	}

	string GetPropertyString(const char* lpDestination, const char* lpPath,
		const char* lpInterface, const char* lpMember)
	{
		string str;

		if(m_pBus != nullptr)
		{
			char* lpValue = nullptr;

			if(sd_bus_get_property_string(m_pBus, lpDestination, lpPath,
				lpInterface, lpMember, nullptr, &lpValue) >= 0)
			{
				if(lpValue != nullptr) str = lpValue;
			}

			if(lpValue != nullptr) free(lpValue);
		}

		return str;
	}

private:
	sd_bus* m_pBus;
};

// UInput-based sending of input events, using libxkbcommon for
// translating characters to keypresses.
// https://www.kernel.org/doc/html/latest/input/uinput.html
// https://xkbcommon.org/doc/current/
class CUInputDevice
{
private:
	CUInputDevice(const uint16_t* pKeys, uint32_t cKeys) :
		m_fd(-1), m_bDeviceCreated(false), m_tLastEvent(0)
	{
		if(pKeys == nullptr) return;

		m_fd = open("/dev/uinput", O_WRONLY | O_NONBLOCK);
		if(m_fd < 0) return;

		ioctl(m_fd, UI_SET_EVBIT, EV_SYN);
		if(ioctl(m_fd, UI_SET_EVBIT, EV_KEY) < 0) return;

		// When simply registering all keys from 1 to KEY_MAX, UInput
		// refuses to send any key on some systems (e.g. Ubuntu 21.04);
		// therefore, we register only the keys that we really need
		for(uint32_t i = 0; i < cKeys; ++i)
		{
			const uint16_t uKey = pKeys[i];
			if((uKey == 0) || (uKey >= KEY_MAX)) continue;
			if(m_sKeys.find(uKey) != m_sKeys.end()) continue;

			if(ioctl(m_fd, UI_SET_KEYBIT, static_cast<size_t>(uKey)) >= 0)
				m_sKeys.insert(uKey);
		}

		uinput_setup s;
		memset(&s, 0, sizeof(uinput_setup));
		s.id.bustype = BUS_VIRTUAL;
		s.id.vendor = 1;
		s.id.product = 1;
		s.id.version = 1;
		strncpy(s.name, "KeePass Auto-Type", UINPUT_MAX_NAME_SIZE - 1);

		if(ioctl(m_fd, UI_DEV_SETUP, &s) < 0) return;
		if(ioctl(m_fd, UI_DEV_CREATE) < 0) return;
		sleep(1); // Allow detect./init. of the device; see UInput doc.
		m_bDeviceCreated = true;
	}

public:
	virtual ~CUInputDevice()
	{
		if(m_bDeviceCreated)
		{
			sleep(1); // Allow proc. the last events; see UInput doc.
			ioctl(m_fd, UI_DEV_DESTROY);

			m_bDeviceCreated = false;
		}

		if(m_fd >= 0)
		{
			close(m_fd);

			m_fd = -1;
		}
	}

	static CUInputDevice* CreateInstance(const uint16_t* pKeys, uint32_t cKeys)
	{
		CUInputDevice* p = new CUInputDevice(pKeys, cKeys);
		if(!p->m_bDeviceCreated) { delete p; return nullptr; }
		return p;
	}

	void StartSequence()
	{
		UpdateCharKeyMap();
	}

	void EndSequence() const
	{
	}

	void SendKey(uint16_t uKey)
	{
		SendKey(uKey, true);
		SendKey(uKey, false);
	}

	void SendKey(uint16_t uKey, bool bDown)
	{
		if(!m_bDeviceCreated) return;

		const int64_t tDiff = GetTimeMSec() - m_tLastEvent;
		if((tDiff >= 0) && (tDiff < 33)) SleepMSec(33 - tDiff);

		input_event ie;
		memset(&ie, 0, sizeof(input_event));

		ie.type = EV_KEY;
		ie.code = uKey;
		ie.value = (bDown ? 1 : 0);
		write(m_fd, &ie, sizeof(input_event));

		ie.type = EV_SYN;
		ie.code = SYN_REPORT;
		ie.value = 0;
		write(m_fd, &ie, sizeof(input_event));

		m_tLastEvent = GetTimeMSec();
	}

	void SendChar(uint32_t ch)
	{
		unordered_map<uint32_t, uint32_t>::const_iterator it = m_mCharToKeys.find(ch);
		if(it == m_mCharToKeys.end())
		{
			SendCharByComposition(ch);
			return;
		}

		const uint32_t uKeys = it->second;

		// Shift should be the first;
		// https://www.linux.org/threads/alt-keys-and-linux.11517/
		if((uKeys & KPUI_K_MOD_SHIFT_L) != 0)
			SendKey(KEY_LEFTSHIFT, true);
		if((uKeys & KPUI_K_MOD_SHIFT_R) != 0)
			SendKey(KEY_RIGHTSHIFT, true);
		if((uKeys & KPUI_K_MOD_CTRL_L) != 0)
			SendKey(KEY_LEFTCTRL, true);
		if((uKeys & KPUI_K_MOD_CTRL_R) != 0)
			SendKey(KEY_RIGHTCTRL, true);
		if((uKeys & KPUI_K_MOD_ALT_L) != 0)
			SendKey(KEY_LEFTALT, true);
		if((uKeys & KPUI_K_MOD_ALT_R) != 0)
			SendKey(KEY_RIGHTALT, true);

		SendKey(static_cast<uint16_t>(uKeys & KPUI_K_CODE));

		if((uKeys & KPUI_K_MOD_SHIFT_L) != 0)
			SendKey(KEY_LEFTSHIFT, false);
		if((uKeys & KPUI_K_MOD_SHIFT_R) != 0)
			SendKey(KEY_RIGHTSHIFT, false);
		if((uKeys & KPUI_K_MOD_CTRL_L) != 0)
			SendKey(KEY_LEFTCTRL, false);
		if((uKeys & KPUI_K_MOD_CTRL_R) != 0)
			SendKey(KEY_RIGHTCTRL, false);
		if((uKeys & KPUI_K_MOD_ALT_L) != 0)
			SendKey(KEY_LEFTALT, false);
		if((uKeys & KPUI_K_MOD_ALT_R) != 0)
			SendKey(KEY_RIGHTALT, false);
	}

	void SetProperty(const char* lpKey, const char* lpValue)
	{
		if((lpKey == nullptr) || (*lpKey == '\0')) return;

		const string strKey(lpKey);

		if(lpValue == nullptr)
		{
			unordered_map<string, string>::const_iterator it = m_mProps.find(strKey);
			if(it != m_mProps.end()) m_mProps.erase(it);
		}
		else
		{
			const string strValue(lpValue);
			m_mProps[strKey] = strValue;
		}
	}

	string GetProperty(const char* lpKey, const char* lpDefault = nullptr) const
	{
		string strDefault;
		if((lpDefault != nullptr) && (*lpDefault != '\0')) strDefault = lpDefault;

		if((lpKey == nullptr) || (*lpKey == '\0')) return strDefault;

		const string strKey(lpKey);
		unordered_map<string, string>::const_iterator it = m_mProps.find(strKey);
		if(it != m_mProps.end()) return it->second;
		return strDefault;
	}

private:
	bool UpdateCharKeyMap()
	{
		bool bResult = false;

		m_mCharToKeys.clear();

		xkb_rule_names rn;
		memset(&rn, 0, sizeof(xkb_rule_names));
		CSdBus sdb;
		const char* lpDest = "org.freedesktop.locale1";
		const char* lpPath = "/org/freedesktop/locale1";
		const char* lpInterface = "org.freedesktop.locale1";
		const string strModel = sdb.GetPropertyString(lpDest, lpPath,
			lpInterface, "X11Model");
		if(strModel.size() != 0) rn.model = strModel.c_str();
		const string strLayout = sdb.GetPropertyString(lpDest, lpPath,
			lpInterface, "X11Layout");
		if(strLayout.size() != 0) rn.layout = strLayout.c_str();
		const string strVariant = sdb.GetPropertyString(lpDest, lpPath,
			lpInterface, "X11Variant");
		if(strVariant.size() != 0) rn.variant = strVariant.c_str();
		const string strOptions = sdb.GetPropertyString(lpDest, lpPath,
			lpInterface, "X11Options");
		if(strOptions.size() != 0) rn.options = strOptions.c_str();

		xkb_context* pCtx = xkb_context_new(XKB_CONTEXT_NO_FLAGS);
		if(pCtx != nullptr)
		{
			xkb_keymap* pKeyMap = xkb_keymap_new_from_names(pCtx, &rn,
				XKB_KEYMAP_COMPILE_NO_FLAGS);
			if(pKeyMap != nullptr)
			{
				xkb_state* pState = xkb_state_new(pKeyMap);
				if(pState != nullptr)
				{
					bResult = UpdateCharKeyMapXkb(pKeyMap, pState);

					xkb_state_unref(pState);
				}

				xkb_keymap_unref(pKeyMap);
			}

			xkb_context_unref(pCtx);
		}

		return bResult;
	}

	bool UpdateCharKeyMapXkb(xkb_keymap* pKeyMap, xkb_state* pState)
	{
		const size_t cMods = 4;
		const uint32_t vMods[cMods] = {
			// https://help.ubuntu.com/community/ComposeKey
			KPUI_K_MOD_ALT_R | KPUI_K_MOD_SHIFT_L,
			KPUI_K_MOD_ALT_R,
			KPUI_K_MOD_SHIFT_L,
			0
		};

		const xkb_keycode_t xkcMin = KeyClampX(xkb_keymap_min_keycode(pKeyMap));
		const xkb_keycode_t xkcMax = KeyClampX(xkb_keymap_max_keycode(pKeyMap));

		for(size_t uMods = 0; uMods < cMods; ++uMods)
		{
			const uint32_t m = vMods[uMods];

			if((m & KPUI_K_MOD_SHIFT_L) != 0)
				xkb_state_update_key(pState, KeyLToX(KEY_LEFTSHIFT), XKB_KEY_DOWN);
			if((m & KPUI_K_MOD_SHIFT_R) != 0)
				xkb_state_update_key(pState, KeyLToX(KEY_RIGHTSHIFT), XKB_KEY_DOWN);
			if((m & KPUI_K_MOD_CTRL_L) != 0)
				xkb_state_update_key(pState, KeyLToX(KEY_LEFTCTRL), XKB_KEY_DOWN);
			if((m & KPUI_K_MOD_CTRL_R) != 0)
				xkb_state_update_key(pState, KeyLToX(KEY_RIGHTCTRL), XKB_KEY_DOWN);
			if((m & KPUI_K_MOD_ALT_L) != 0)
				xkb_state_update_key(pState, KeyLToX(KEY_LEFTALT), XKB_KEY_DOWN);
			if((m & KPUI_K_MOD_ALT_R) != 0)
				xkb_state_update_key(pState, KeyLToX(KEY_RIGHTALT), XKB_KEY_DOWN);

			for(xkb_keycode_t xkc = xkcMin; xkc <= xkcMax; ++xkc)
			{
				const uint16_t k = KeyXToL(xkc);
				if(m_sKeys.find(k) == m_sKeys.end()) continue;

				const uint32_t ch = xkb_state_key_get_utf32(pState, xkc);
				if(ch == 0) continue;

				m_mCharToKeys[ch] = m | k;
			}

			if((m & KPUI_K_MOD_SHIFT_L) != 0)
				xkb_state_update_key(pState, KeyLToX(KEY_LEFTSHIFT), XKB_KEY_UP);
			if((m & KPUI_K_MOD_SHIFT_R) != 0)
				xkb_state_update_key(pState, KeyLToX(KEY_RIGHTSHIFT), XKB_KEY_UP);
			if((m & KPUI_K_MOD_CTRL_L) != 0)
				xkb_state_update_key(pState, KeyLToX(KEY_LEFTCTRL), XKB_KEY_UP);
			if((m & KPUI_K_MOD_CTRL_R) != 0)
				xkb_state_update_key(pState, KeyLToX(KEY_RIGHTCTRL), XKB_KEY_UP);
			if((m & KPUI_K_MOD_ALT_L) != 0)
				xkb_state_update_key(pState, KeyLToX(KEY_LEFTALT), XKB_KEY_UP);
			if((m & KPUI_K_MOD_ALT_R) != 0)
				xkb_state_update_key(pState, KeyLToX(KEY_RIGHTALT), XKB_KEY_UP);
		}

		return (m_mCharToKeys.size() != 0);
	}

	static xkb_keysym_t KeyLToX(uint16_t uKey)
	{
		return (static_cast<xkb_keysym_t>(uKey) + 8);
	}

	static uint16_t KeyXToL(xkb_keysym_t xKey)
	{
		return static_cast<uint16_t>(xKey - 8);
	}

	static xkb_keysym_t KeyClampX(xkb_keysym_t xKey)
	{
		return min(max(xKey, 8U), 0xFFFFU + 8U);
	}

	// https://help.ubuntu.com/community/ComposeKey
	// https://en.wikipedia.org/wiki/Alt_code
	bool SendCharByComposition(uint32_t ch)
	{
		const char* lpReq = "0123456789abcdefu";
		for(int i = 0; lpReq[i] != '\0'; ++i)
		{
			if(m_mCharToKeys.find(static_cast<unsigned char>(lpReq[i])) ==
				m_mCharToKeys.end())
				return false;
		}

		char sz[16] = { '\0' };
		sprintf(sz, "%04x", static_cast<unsigned int>(ch));

		SendKey(KEY_LEFTSHIFT, true);
		SendKey(KEY_LEFTCTRL, true);
		SendChar('u');
		SendKey(KEY_LEFTCTRL, false);
		SendKey(KEY_LEFTSHIFT, false);

		for(int i = 0; sz[i] != '\0'; ++i)
			SendChar(static_cast<unsigned char>(sz[i]));

		SendKey(KEY_ENTER);

		return true;
	}

	static int64_t GetTimeMSec()
	{
		timespec t;
		if(clock_gettime(CLOCK_REALTIME, &t) < 0) return 0;

		return ((static_cast<int64_t>(t.tv_sec) * 1000) +
			static_cast<int64_t>(t.tv_nsec / (1000 * 1000)));
	}

	static void SleepMSec(int64_t ms)
	{
		if(ms < 0) ms = 0;

		timespec t;
		memset(&t, 0, sizeof(timespec));
		t.tv_sec = static_cast<time_t>(ms / 1000);
		t.tv_nsec = static_cast<long>(ms % 1000) * 1000 * 1000;

		nanosleep(&t, nullptr);
	}

	int m_fd;
	bool m_bDeviceCreated;

	unordered_set<uint16_t> m_sKeys;
	unordered_map<uint32_t, uint32_t> m_mCharToKeys;
	unordered_map<string, string> m_mProps;

	int64_t m_tLastEvent;
};

DLL_PUBLIC_C void* KpuiOpenDevice(const uint16_t* pKeys, uint32_t cKeys)
{
	try { return CUInputDevice::CreateInstance(pKeys, cKeys); }
	catch(...) { }
	return nullptr;
}

DLL_PUBLIC_C void KpuiCloseDevice(void* hDevice)
{
	if(hDevice == nullptr) return;

	try
	{
		CUInputDevice* p = reinterpret_cast<CUInputDevice*>(hDevice);
		delete p;
	}
	catch(...) { }
}

DLL_PUBLIC_C void KpuiStartSequence(void* hDevice)
{
	if(hDevice == nullptr) return;

	try
	{
		CUInputDevice* p = reinterpret_cast<CUInputDevice*>(hDevice);
		p->StartSequence();
	}
	catch(...) { }
}

DLL_PUBLIC_C void KpuiEndSequence(void* hDevice)
{
	if(hDevice == nullptr) return;

	try
	{
		CUInputDevice* p = reinterpret_cast<CUInputDevice*>(hDevice);
		p->EndSequence();
	}
	catch(...) { }
}

DLL_PUBLIC_C void KpuiSendKey(void* hDevice, uint16_t uKey, uint32_t uFlags)
{
	if(hDevice == nullptr) return;

	try
	{
		CUInputDevice* p = reinterpret_cast<CUInputDevice*>(hDevice);

		if((uFlags & KPUI_S_DOWN) != 0)
			p->SendKey(uKey, true);
		if((uFlags & KPUI_S_UP) != 0)
			p->SendKey(uKey, false);
	}
	catch(...) { }
}

DLL_PUBLIC_C void KpuiSendChar(void* hDevice, uint32_t ch)
{
	if(hDevice == nullptr) return;

	try
	{
		CUInputDevice* p = reinterpret_cast<CUInputDevice*>(hDevice);
		p->SendChar(ch);
	}
	catch(...) { }
}

DLL_PUBLIC_C void KpuiSetProperty(void* hDevice, const char* lpKey, const char* lpValue)
{
	if(hDevice == nullptr) return;

	try
	{
		CUInputDevice* p = reinterpret_cast<CUInputDevice*>(hDevice);
		p->SetProperty(lpKey, lpValue);
	}
	catch(...) { }
}
